﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Principal; // TODO Why isn't this being recognized? Is it due to the static class or method declaration?
using System.Web;
using System.Web.SessionState;
using AuthLiteTokenProfileManager.Code.Constants;

namespace AuthLiteTokenProfileManager.Code.Helpers
{
	public static class SessionHelper
	{
		/// <summary>
		/// Ensures the Session is still valid based on custom business requirements.
		/// </summary>
		/// <returns></returns>
		public static bool IsValid()
		{
			bool valid = false;

			HttpSessionState s = HttpContext.Current.Session;			
            if (s[Constants.Constants.USERNAME] != null && 
				s[Constants.Constants.DOMAIN] != null)
			{
				valid = true;
			}

			return valid;
		}

        public static string Domain
        {
            get { return HttpContext.Current.Session[Constants.Constants.DOMAIN].ToString(); }
        }
        public static string Username
        {
            get { return HttpContext.Current.Session[Constants.Constants.USERNAME].ToString(); }
        }
        public static string Password
        {
            get { return HttpContext.Current.Session[Constants.Constants.PASSWORD].ToString(); }
            set { HttpContext.Current.Session[Constants.Constants.PASSWORD] = value; }
        }
        public static string CacheOtp // during pass change, the initial OTP provided is still fresh, need to use after pass change
        {
            get { return (string)HttpContext.Current.Session[Constants.Constants.CACHEOTP]; }
            set { HttpContext.Current.Session[Constants.Constants.CACHEOTP] = value; }
        }
        public static WindowsIdentity WindowsIdentity
        {
            get { return HttpContext.Current.Session[Constants.Constants.WINDOWS_IDENTITY] as WindowsIdentity; }
            set { HttpContext.Current.Session[Constants.Constants.WINDOWS_IDENTITY] = value; }
        }
        public static string OATHDescription
        {
            get { return HttpContext.Current.Session[Constants.Constants.OATH_DESCRIPTION].ToString(); }
            set { HttpContext.Current.Session[Constants.Constants.OATH_DESCRIPTION] = value; }
        }
        public static string OATHIssuer
        {
            get { return HttpContext.Current.Session[Constants.Constants.OATH_ISSUER].ToString(); }
            set { HttpContext.Current.Session[Constants.Constants.OATH_ISSUER] = value; }
        }
        public static byte[] OATHSecret
        {
            get { return HttpContext.Current.Session[Constants.Constants.OATH_SECRET] as byte[]; }
            set { HttpContext.Current.Session[Constants.Constants.OATH_SECRET] = value; }
        }
        public static bool OATHOnline
        {
            get { return HttpContext.Current.Session[Constants.Constants.OATH_ONLINE] == "true"; }
            set { HttpContext.Current.Session[Constants.Constants.OATH_ONLINE] = value? "true" : "false"; }
        }
        public static bool WasAuthLiteUser
        {
            get { return HttpContext.Current.Session[Constants.Constants.WAS_AUTHLITE_USER] == "true"; }
            set { HttpContext.Current.Session[Constants.Constants.WAS_AUTHLITE_USER] = value ? "true" : "false"; }
        }
		
		/// <summary>
		/// End the current Session.
		/// </summary>
		public static void End()
		{
			HttpSessionState s = HttpContext.Current.Session;
            if (WindowsIdentity != null) WindowsIdentity.Dispose();
			s.Clear(); // Remove values immediately (overkill?)
			s.Abandon(); // Destroy Session at the end of the current request
		}
	}
}