﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Web.Mvc;
using AuthLiteTokenProfileManager.Code.Constants;
using AuthLiteTokenProfileManager.Code.Helpers;
using log4net;

//http://stackoverflow.com/questions/3214774/how-to-redirect-from-onactionexecuting-in-base-controller

namespace AuthLiteTokenProfileManager.Controllers
{
    public abstract class BaseController : Controller
    {

        private static readonly ILog Log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

		// Ensure Session Is Valid before each action/step is performed, otherwise bump the user back to the home screen
		protected override void OnActionExecuting(ActionExecutingContext filterContext)
		{
			
			// Bypass check on Home controller because this is where sessions are started.
			if (filterContext.ActionDescriptor.ControllerDescriptor.ControllerName.ToLower() == Constants.CONTROLLER_MAIN)
			{
				base.OnActionExecuting(filterContext);
				return;
			} 

			if(SessionHelper.IsValid())
			{
				base.OnActionExecuting(filterContext);
			} else
			{
                Log.Info("Session is invalid. Redirecting to Home SessionExpired Action. ");
				SessionHelper.End();				
				filterContext.Result = RedirectToAction("SessionExpired", "Home");
			}			
		}
	}
}