﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Web.Mvc;
using AuthLiteTokenProfileManager.Models;
using AuthLiteTokenProfileManager.Code.Helpers;
using System.Security.Principal;
using System.ComponentModel;
using AuthLiteTokenProfileManager.Code;
using Collective.AuthLite.Protocol;
using Microsoft.Win32;
using log4net;
using TPMCommon;

namespace AuthLiteTokenProfileManager.Controllers
{
	public class LoginController : BaseController
	{
        private static readonly ILog Log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        private string GetUserString()
        {
            return string.Format(@"{0}\{1}", SessionHelper.Domain, SessionHelper.Username);
        }
		public ActionResult Index()
		{
            try
            {
                if (SettingsHelper.SanityCheckS4UCache)
                {
                    using (RegistryKey key = Registry.LocalMachine.OpenSubKey(@"System\CurrentControlSet\Control\Lsa\Kerberos\Parameters"))
                    {
                        if (key == null)
                        {
                            string msg = "Error: S4UCache is not set to 0";
                            ModelState.AddModelError("S4UCache", msg);
                            Log.Warn("S4UCache " + msg);
                        }
                        else
                        {
                            object o = key.GetValue("CacheS4UTickets");
                            if (o == null)
                            {
                                string msg = "Error: S4UCache is not set to 0";
                                ModelState.AddModelError("S4UCache", msg);
                                Log.Warn("S4UCache " + msg);
                            }
                            else
                            {
                                int v = (int)key.GetValue("CacheS4UTickets");
                                if (v != 0)
                                {
                                    string msg = "Error: S4UCache is not set to 0";
                                    ModelState.AddModelError("S4UCache", msg);
                                    Log.Warn("S4UCache " + msg);
                                }
                            }
                        }
                    }
                }
                if (AuthLiteOperations.Instance.IsAuthLiteUser(SessionHelper.Domain, SessionHelper.Username))
                {
                    SessionHelper.WasAuthLiteUser = true;
                    if (SettingsHelper.AlwaysUse1FLogons)
                    {
                        LoginPasswordFormModel m = new LoginPasswordFormModel();
                        m.User = GetUserString();
                        Log.Info("WasAuthLiteUser TRUE (AlwaysUse1FLogons TRUE). Returning Login view.");
                        return View("Login", m);
                    }
                    else
                    {
                        LoginPasswordAndOtpFormModel m = new LoginPasswordAndOtpFormModel();
                        m.User = GetUserString();
                        Log.Info("WasAuthLiteUser FALSE (AlwaysUse1FLogons FALSE). Returning LoginWithOtp view.");
                        return View("LoginWithOtp", m);
                    }
                }
                else
                {
                    SessionHelper.WasAuthLiteUser = false;
                    LoginPasswordFormModel m = new LoginPasswordFormModel();
                    m.User = GetUserString();
                    Log.Info("WasAuthLiteUser FALSE. Return Login view.");
                    return View("Login", m);
                }
            }
            catch (Win32Exception e)
            {
                Log.Error("Error during Login Index Action. ", e);
                ModelState.AddModelError("IsAuthLiteUserCheck", string.Format("Could not look up user. [Win32 Error {0}, {1}]", e.ErrorCode, e.NativeErrorCode));
                return View("Login", new LoginPasswordFormModel());
            }
            
		}

        public ActionResult NowAuthLiteUser()
        {
            // I *should* be able to do this, but for some reason it ruins the TempData
            // (I did have it showing ok when I returned a View instead of doing a Redirect,
            // but then the Submit button on the form wouldn't go to the right action even after I 
            // explicitly made BeginForm("Index")
            // Anyway, I just made the caller Dispose the WindowsIdentity, and that's pretty well
            // good enough to be safe.
            //SessionHelper.End();

            TempData["success"] = "You are now an AuthLite user. Please log in again using 2-factor.";
            Log.Info("NowAuthLiteUser. ");
            return RedirectToAction("Index");
        }

		[HttpPost]
		[ValidateAntiForgeryToken]
		public ActionResult Login(LoginPasswordFormModel m)
		{
			if (!ModelState.IsValid)
			{
                Log.Info("Invalid Login ModelState. Return to current view. ");
				return View(m);
			}
            m.User = GetUserString();
            string password = m.Password.ActiveDirectoryPassword;
            try
            {
                WindowsIdentity wi = IdentityHelper.Instance.LogonUser(SessionHelper.Domain, SessionHelper.Username, password);
                if(wi == null)
                {
                    Exception ex = new Exception("LogonUser returned null");                    
                    Log.Error("Login WindowsIndentity Error. ", ex);
                    throw ex;
                }
                SessionHelper.WindowsIdentity = wi;
                SessionHelper.Password = password;
                Log.Info("Login successful. Redirecting to Manager Index Action. ");
				return RedirectToAction("Index", "Manager");
            }
            catch(Win32Exception e)
            {
                string detail = "";
                switch(e.NativeErrorCode)
                {
                    case 1326:
                        detail = ": Wrong password?";
                        break;
                    case 1330: //ERROR_PASSWORD_EXPIRED
                    case 1907: //ERROR_PASSWORD_MUST_CHANGE
                        SessionHelper.Password = password; // to use after reset
                        return View("NewPassword");
                        break;
                }
                string msg = string.Format("Could not log on, please try again. [Error {0}, {1}{2}]", e.ErrorCode, e.NativeErrorCode, detail);
                ModelState.AddModelError("LogonUserFailed", msg);
                Log.Warn("LogonUserFailed " + msg); 
				return View(m);
            }
            catch(Exception e)
            {
                string msg = string.Format("Could not log on. [Error {0}]", e.Message);
                ModelState.AddModelError("GeneralError", msg);
                Log.Warn("GeneralError " + msg);
                return View(m);
            }		
		}

        [HttpPost]
        [ValidateAntiForgeryToken]
        public ActionResult LoginWithOtp(LoginPasswordAndOtpFormModel m)
        {
            if (!ModelState.IsValid)
            {
                m.Passcode = string.Empty;
                ModelState.Remove("Passcode");
                Log.Info("LoginWithOtp form ModelState is invalid. Returning to current view. ");
                return View(m);
            }
            m.User = GetUserString();
            string password = m.Password.ActiveDirectoryPassword;
            string passcode = m.Passcode;
            m.Passcode = string.Empty;
            ModelState.Remove("Passcode");
            try
            {
                WindowsIdentity wi = IdentityHelper.Instance.LogonUser(SessionHelper.Domain, SessionHelper.Username, password + "-" + passcode);
                if (wi == null)
                {
                    Exception ex = new Exception("LogonUser returned null");
                    Log.Error("LoginWithOtp WindowsIndentity Error. ", ex);
                    throw ex;
                }
                using(new Impersonation(wi))
                {
                    AuthLiteIsAuthLiteUserStatus status = AuthLiteOperations.Instance.IsAuthLiteUser(SessionHelper.Domain);
                    if(status != AuthLiteIsAuthLiteUserStatus.Yes2FA)
                    {
                        string msg = string.Format("This logon session should have been 2-factor, but was not. [status {0}]", status);
                        ModelState.AddModelError("LogonUserNot2F", msg);
                        Log.Warn("LogonUserNot2F " + msg);
                        return View(m);
                    }
                }
                SessionHelper.WindowsIdentity = wi;
                SessionHelper.Password = password;

                Log.Info("LoginWithOtp successful. Redirecting to Manager Index Action. ");
                return RedirectToAction("Index", "Manager");
            }
            catch (Win32Exception e)
            {
                string detail = "";
                switch (e.NativeErrorCode)
                {
                    case 1326:
                        detail = ": Wrong password or OTP?";
                        break;
                    case 1327:
                    case 1385:
                        detail = ": Wrong One-time Passcode?";
                        break;
                    case 1330: //ERROR_PASSWORD_EXPIRED
                    case 1907: //ERROR_PASSWORD_MUST_CHANGE
                        SessionHelper.Password = password; // to use after reset
                        SessionHelper.CacheOtp = passcode;
                        return View("NewPassword");
                        break;
                }
                string msg = string.Format("Could not log on, please try again. [Error {0}, {1}{2}]", e.ErrorCode, e.NativeErrorCode, detail);
                ModelState.AddModelError("LogonUserFailed", msg);
                Log.Warn("LoginWithOtp LogonUserFailed " + msg);
                return View(m);
            }
            catch (Exception e)
            {
                string msg = string.Format("Could not log on. [Error {0}]", e.Message);
                ModelState.AddModelError("GeneralError", msg);
                Log.Warn("GeneralError " + msg);
                return View(m);
            }
        }

		[HttpPost]
		[ValidateAntiForgeryToken]
		public ActionResult NewPassword(NewPasswordFormModel m)
		{
			if (!ModelState.IsValid)
			{
                Log.Info("Invalid NewPassword ModelState. Return to current view. ");
                m.NewPassword = string.Empty;
                ModelState.Remove("NewPassword");
                m.ConfirmPassword = string.Empty;
                ModelState.Remove("ConfirmPassword");
				return View(m);
			}
            try
            {
                IdentityHelper.Instance.ChangePassword(SessionHelper.Domain, SessionHelper.Username, SessionHelper.Password, m.NewPassword);
                SessionHelper.Password = m.NewPassword; // so the following logon action will work
                Log.Info("Password change successful. Logging in");
                return LoginAfterNewPassword();
            }
            catch(Win32Exception e)
            {
                string detail = "";
                string msg = string.Format("Could not change password, please try again. [Error: {0}, {1}{2}]", e.ErrorCode, e.NativeErrorCode, detail);
                ModelState.AddModelError("ChangePasswordFailed", msg);
                Log.Warn("ChangePasswordFailed " + msg);
                m.NewPassword = string.Empty;
                ModelState.Remove("NewPassword");
                m.ConfirmPassword = string.Empty;
                ModelState.Remove("ConfirmPassword");
                return View(m);
            }
            catch(Exception e)
            {
                string msg = string.Format("Could not change password. [Error: {0}]", e.Message);
                ModelState.AddModelError("ChangePasswordFailed", msg);
                Log.Warn("GeneralError " + msg);
                m.NewPassword = string.Empty;
                ModelState.Remove("NewPassword");
                m.ConfirmPassword = string.Empty;
                ModelState.Remove("ConfirmPassword");
                return View(m);
            }		
		}

        // I tried making this a view and RedirectToAction'ing to it after NewPassword, but it 404'd. maybe needed a view with form or something?
        // anyway we already posted and don't need to collect more info, so just calling it directly now
        public ActionResult LoginAfterNewPassword()
        {
            bool expect2f = false;
            try
            {
                string passwordAndOtp = SessionHelper.Password;
                if(!string.IsNullOrEmpty(SessionHelper.CacheOtp))
                {
                    passwordAndOtp += "-" + SessionHelper.CacheOtp;
                    expect2f = true;
                    SessionHelper.CacheOtp = string.Empty;
                }
                WindowsIdentity wi = IdentityHelper.Instance.LogonUser(SessionHelper.Domain, SessionHelper.Username, passwordAndOtp);
                if (wi == null)
                {
                    Exception ex = new Exception("LogonUser returned null");
                    Log.Error("LoginAfterNewPassword WindowsIndentity Error. ", ex);
                    throw ex;
                }
                using (new Impersonation(wi))
                {
                    AuthLiteIsAuthLiteUserStatus status = AuthLiteOperations.Instance.IsAuthLiteUser(SessionHelper.Domain);
                    if (expect2f && status != AuthLiteIsAuthLiteUserStatus.Yes2FA)
                    {
                        string msg = string.Format("This logon session should have been 2-factor, but was not. [status {0}]", status);
                        TempData["error"] = msg;
                        Log.Warn("LogonUserNot2F " + msg);
                        return RedirectToAction("Index");
                    }
                }
                SessionHelper.WindowsIdentity = wi;

                Log.Info("LoginAfterNewPassword successful. Redirecting to Manager Index Action. ");
                return RedirectToAction("Index", "Manager");
            }
            catch (Win32Exception e)
            {
                string detail = "";
                switch (e.NativeErrorCode)
                {
                    case 1326:
                        detail = ": Wrong password or OTP?";
                        break;
                    case 1327:
                    case 1385:
                        detail = ": Wrong One-time Passcode?";
                        break;
                }
                string msg = string.Format("Could not log on, please try again. [Error {0}, {1}{2}]", e.ErrorCode, e.NativeErrorCode, detail);
                TempData["error"] = msg;
                Log.Warn("LoginAfterNewPassword LogonUserFailed " + msg);
                return RedirectToAction("Index");
            }
            catch (Exception e)
            {
                string msg = string.Format("Could not log on. [Error {0}]", e.Message);
                TempData["error"] = msg;
                Log.Warn("GeneralError " + msg);
                return RedirectToAction("Index");
            }
        }
		
		public ActionResult Logout()
		{
            // calls WindowsIdentity.Dispose
			SessionHelper.End();
            Log.Info("Logout. Redirecting to Home Index Action. ");
			return RedirectToAction("Index", "Home");
		}
	}
}