﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Net;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Web.Mvc;
using AuthLiteTokenProfileManager.Models;
using AuthLiteTokenProfileManager.Code.Constants;
using Collective.AuthLite.Protocol;

using ThoughtWorks.QRCode.Codec;
using AuthLiteTokenProfileManager.Code.Helpers;
using Collective.Util;
using System.Security.Principal;
using AuthLiteTokenProfileManager.Code;
using System.ComponentModel;
using Collective.Security;
using System.Web.Configuration;
using log4net;
using Collective.AuthLite.Util;
using TPMCommon;

namespace AuthLiteTokenProfileManager.Controllers
{
    public class ManagerController : BaseController
    {
        
        private static readonly ILog Log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        public ActionResult Index()
        {
            if(!SessionHelper.WasAuthLiteUser && AuthLiteOperations.Instance.IsAuthLiteUser(SessionHelper.Domain, SessionHelper.Username))
            {
                // state changed, old session goes away because it's not good enough to log in any more
                SessionHelper.WindowsIdentity.Dispose();
                if (SettingsHelper.AfterBootstrapGoto == Constants.CONTROLLER_LOGIN)
                {
                    return RedirectToAction("NowAuthLiteUser", "Login");
                }
                else
                {
                    return RedirectToAction("NowAuthLiteUser", "Home");
                }
            }
            ManagerModel m = new ManagerModel();
            m.PinEnabled = SettingsHelper.UserPinEnabled && SessionHelper.WasAuthLiteUser;
            m.TokensListEnabled = SessionHelper.WasAuthLiteUser;
            return View("Manager", m);
        }

		/// <summary>
		/// Display the Add YubiKey screen
		/// </summary>
		/// <returns></returns>
		public ActionResult AddYubiKey()
		{
            if (!SessionHelper.WasAuthLiteUser && SettingsHelper.UserPinEnabled)
            {
                AddYubiKeyAndPinFormModel m = new AddYubiKeyAndPinFormModel();
                return View("AddYubiKeyAndPin", m);
            }
            else
            {
                AddYubiKeyFormModel m = new AddYubiKeyFormModel();
                return View("AddYubiKey", m);
            }
		}

		/// <summary>
		/// Handle the Add YubiKey and pin screen form submission
		/// </summary>
		/// <param name="m"></param>
		/// <returns></returns>
		[HttpPost]
		[ValidateAntiForgeryToken]
		public ActionResult AddYubiKeyAndPin(AddYubiKeyAndPinFormModel m)
        {
            if (!ModelState.IsValid)
            {
                m.YubiKey.YubikeyPasscode = string.Empty;
                ModelState.Remove("YubiKey.YubikeyPasscode");

                return View(m);
            }

            try
            {
                using (new Impersonation(SessionHelper.WindowsIdentity))
                {
                    try
                    {
                        AuthLiteOperations.Instance.SetUserPin(SessionHelper.Domain, SessionHelper.Username, m.Pin.Pin);
                    }
                    catch (Win32Exception e)
                    {
                        ModelState.AddModelError("SetUserPinFailed", string.Format("Could not set the user PIN. [Win32 Error {0}, {1}]", e.ErrorCode, e.NativeErrorCode));
                        m.YubiKey.YubikeyPasscode = string.Empty;
                        ModelState.Remove("YubiKey.YubikeyPasscode");
                        return View(m);
                    }
                    catch (Exception e)
                    {
                        ModelState.AddModelError("SetUserPinFailed", string.Format("Could not set the user PIN. [General Error {0}]", e.Message));
                        m.YubiKey.YubikeyPasscode = string.Empty;
                        ModelState.Remove("YubiKey.YubikeyPasscode");
                        return View(m);
                    }
                }
            }
            catch (Win32Exception e)
            {
                ModelState.AddModelError("ImpersonationFailed", string.Format("Could not use logon token. [Error {0}, {1}]", e.ErrorCode, e.NativeErrorCode));
                m.YubiKey.YubikeyPasscode = string.Empty;
                ModelState.Remove("YubiKey.YubikeyPasscode");
                return View(m);
            }
            catch (Exception e)
            {
                ModelState.AddModelError("GeneralError", string.Format("[General Error {0}]", e.Message));
                m.YubiKey.YubikeyPasscode = string.Empty;
                ModelState.Remove("YubiKey.YubikeyPasscode");
                return View(m);
            }

            return AddYubiKey(m);
        }

		/// <summary>
		/// Handle the Add YubiKey screen form submission
		/// </summary>
		/// <param name="m"></param>
		/// <returns></returns>
		[HttpPost]
		[ValidateAntiForgeryToken]
		public ActionResult AddYubiKey(AddYubiKeyFormModel m)
		{
			if (!ModelState.IsValid)
			{
				m.YubiKey.YubikeyPasscode = string.Empty;
				ModelState.Remove("YubiKey.YubikeyPasscode");

				return View(m);
			}
            string otp = m.YubiKey.YubikeyPasscode;
            // clear the model (this isn't a 'password field' but still want it to be clear each page load)
            m.YubiKey.YubikeyPasscode = string.Empty;
            ModelState.Remove("YubiKey.YubikeyPasscode");

            Impersonation i = null;
            try
            {
                if (SettingsHelper.ProvisionAsUser)
                {
                    i = new Impersonation(SessionHelper.WindowsIdentity);
                }
                    try
                    {
                    AuthLiteOperations.Instance.ProvisionYubiKey(SessionHelper.Domain, SessionHelper.Username, i != null? SessionHelper.Password : null, otp, SettingsHelper.ShouldBootstrap);
                    }
                    catch (Win32Exception e)
                    {
                        string detail = "";
                        switch (e.NativeErrorCode)
                        {
                            case 1326:
                                detail = ": Wrong password?";
                                break;
                        }
                        ModelState.AddModelError("ProvisionTokenFailed", string.Format("Could not assign the token. [Win32 Error {0}, {1}{2}]", e.ErrorCode, e.NativeErrorCode, detail));
                        return View(m);
                    }
                    catch (ProvisionTokenException e)
                    {
                        ModelState.AddModelError("ProvisionTokenFailed", string.Format("Could not assign the token. [Provision Error {0}]", e.Message));
                        return View(m);
                    }
                    catch (Exception e)
                    {
                        ModelState.AddModelError("ProvisionTokenFailed", string.Format("Could not assign the token. [General Error {0}]", e.Message));
                        return View(m);
                    }
                
                }
            catch (Win32Exception e)
            {
                ModelState.AddModelError("ImpersonationFailed", string.Format("Could not use logon token. [Error {0}, {1}]", e.ErrorCode, e.NativeErrorCode));
                return View(m);
            }
            catch (Exception e)
            {
                ModelState.AddModelError("GeneralError", string.Format("[General Error {0}]", e.Message));
                return View(m);
            }
            finally
            {
                if (i != null) i.Dispose();
                i = null;
            }

            TempData["success"] = "YubiKey added successfully.";
            return RedirectToAction("Index", "Manager"); // Redirect to Manager on success
 		}

		/// <summary>
		/// Display the Add OATH Token screen
		/// </summary>
		/// <returns></returns>
        /// 
        [OutputCacheAttribute(VaryByParam = "*", Duration = 0, NoStore = true)]
		public ActionResult AddOATHToken()
		{
            if (!SessionHelper.WasAuthLiteUser && SettingsHelper.UserPinEnabled)
            {
                AddOATHTokenAndPinFormModel m = new AddOATHTokenAndPinFormModel();
                return View("AddOATHTokenAndPin", m);
            }
            else
            {
                AddOATHTokenFormModel m = new AddOATHTokenFormModel();
                return View("AddOATHToken", m);
            }
		}

		/// <summary>
		/// Handles the Add OATH Token and pin screen form submission
		/// </summary>
		/// <param name="m"></param>
		/// <returns></returns>
		[HttpPost]
		[ValidateAntiForgeryToken]
        [OutputCacheAttribute(VaryByParam = "*", Duration = 0, NoStore = true)]
		public ActionResult AddOATHTokenAndPin(AddOATHTokenAndPinFormModel m)
        {
            if(!ModelState.IsValid)
            {
                return View(m);
            }
            Impersonation i = null;
            try
            {
                if (SettingsHelper.ProvisionAsUser)
                {
                    i = new Impersonation(SessionHelper.WindowsIdentity);
                }

                    try
                    {
                        AuthLiteOperations.Instance.SetUserPin(SessionHelper.Domain, SessionHelper.Username, m.Pin.Pin);
                    }
                    catch (Win32Exception e)
                    {
                        ModelState.AddModelError("SetUserPinFailed", string.Format("Could not set the user PIN. [Win32 Error {0}, {1}]", e.ErrorCode, e.NativeErrorCode));
                        return View(m);
                    }
                    catch (Exception e)
                    {
                        ModelState.AddModelError("SetUserPinFailed", string.Format("Could not set the user PIN. [General Error {0}]", e.Message));
                        return View(m);
                    }
                }
            catch (Win32Exception e)
            {
                ModelState.AddModelError("ImpersonationFailed", string.Format("Could not use logon token. [Error {0}, {1}]", e.ErrorCode, e.NativeErrorCode));
                return View(m);
            }
            catch (Exception e)
            {
                ModelState.AddModelError("GeneralError", string.Format("[General Error {0}]", e.Message));
                return View(m);
            }
            finally
            {
                if (i != null) i.Dispose();
                i = null;
            }

            return AddOATHToken(m);
        }

		/// <summary>
		/// Handles the Add OATH Token screen form submission
		/// </summary>
		/// <param name="m"></param>
		/// <returns></returns>
		[HttpPost]
		[ValidateAntiForgeryToken]
        [OutputCacheAttribute(VaryByParam = "*", Duration = 0, NoStore = true)]
		public ActionResult AddOATHToken(AddOATHTokenFormModel m)
		{
			if (!ModelState.IsValid)
			{
				return View(m);
			}
            var oathDigits = AuthLiteOperations.Instance.GetPartitionSetting(SessionHelper.Domain, Collective.AuthLite.Settings.Constants.SETTING_OATH_OTP_LENGTH, "0");
            if(oathDigits != "6" && oathDigits != "8")
            {
                ModelState.AddModelError("ProvisionTokenFailed", "Please change Token Settings -> OATH Digits to 6 or 8 before proceeding");
                return View(m);

            }
            byte[] oathSecret = CryptoUtil.GetRandomBytes(20);
            SessionHelper.OATHSecret = oathSecret;
            string description = string.Format(SettingsHelper.OathDescriptionFormat, SessionHelper.Domain, SessionHelper.Username);
            SessionHelper.OATHDescription = description;
            string issuer = string.Format(SettingsHelper.OathIssuerFormat, SessionHelper.Domain, m.OnlineOffline.Online? "online" : "offline", SessionHelper.Username);
            SessionHelper.OATHIssuer = issuer;
            SessionHelper.OATHOnline = m.OnlineOffline.Online;
			return RedirectToAction("OATHQRCode", "Manager");
		}

        [OutputCacheAttribute(VaryByParam = "*", Duration = 0, NoStore = true)]
		public ActionResult OATHQRCode()
		{
            if (SessionHelper.OATHSecret == null)
            {
                return RedirectToAction("Index");
            }
            else
            {
                return View("OATHQRCode");//no model.
            }
		}

		/// <summary>
		/// Render a QR Image from the available BitmapData
		/// </summary>
		/// <returns></returns>
		public ActionResult RenderQRImage()
		{
			Bitmap image;
            if (SessionHelper.OATHSecret == null)
            {
                //creating bitmap image
                image = new Bitmap(1, 1);
            }
            else
            {
                QRCodeEncoder encoder = new QRCodeEncoder();
                encoder.QRCodeEncodeMode = QRCodeEncoder.ENCODE_MODE.BYTE;
                encoder.QRCodeScale = 3;
                //encoder.QRCodeVersion = 7;
                encoder.QRCodeErrorCorrect = QRCodeEncoder.ERROR_CORRECTION.M;
                var algorithm = (AuthLiteTokenAlgorithm)Enum.Parse(typeof(AuthLiteTokenAlgorithm), AuthLiteOperations.Instance.GetPartitionSetting(SessionHelper.Domain, "DefaultOathAlgorithm", "SHA1") ?? "SHA1", true);
                if (!string.IsNullOrEmpty( SettingsHelper.OathAlgorithm))
                {
                    algorithm = (AuthLiteTokenAlgorithm)Enum.Parse(typeof(AuthLiteTokenAlgorithm), SettingsHelper.OathAlgorithm, true);
                }
                var oathDigits = AuthLiteOperations.Instance.GetPartitionSetting(SessionHelper.Domain, Collective.AuthLite.Settings.Constants.SETTING_OATH_OTP_LENGTH, "0");
                string sdigits = "";
                if (oathDigits == "8")
                {
                    sdigits = "&digits=8";
                }

                string salgo = (algorithm == AuthLiteTokenAlgorithm.SHA1) ? "" : "&algorithm=SHA256";
                string secretEncoded = Collective.Oath.Base32Encoding.ToString(SessionHelper.OATHSecret);
                string url = string.Format("otpauth://totp/{0}?secret={1}&issuer={2}{3}{4}", System.Web.HttpUtility.UrlEncode(SessionHelper.OATHDescription).Replace("+", "%20"),
                    secretEncoded,
                    System.Web.HttpUtility.UrlEncode(SessionHelper.OATHIssuer).Replace("+", "%20"),
                    salgo,
                    sdigits
                    );
                encoder.QRCodeVersion = QRUtil.VersionForString(url);
                image = encoder.Encode(url);
            }
			/*test code // Pull a test image from online
			WebRequest req = WebRequest.Create("http://placekitten.com/400/400");
			WebResponse response = req.GetResponse();
			Stream stream = response.GetResponseStream();
			image = new Bitmap(stream);
            */
			MemoryStream ms = new MemoryStream();
			image.Save(ms, ImageFormat.Png);
			var base64Data = Convert.ToBase64String(ms.ToArray());
			return Content("data:image/gif;base64," + base64Data);
		}

		public ActionResult GetQRBase32String()
		{
			string secretEncoded = Collective.Oath.Base32Encoding.ToString(SessionHelper.OATHSecret);
			return Content(secretEncoded);
		}

        [HttpPost]
		[ValidateAntiForgeryToken]
        public ActionResult QRSuccess()
        {
            byte[] oathSecret = SessionHelper.OATHSecret;
            string description = SessionHelper.OATHDescription;
            SessionHelper.OATHSecret = null;
            SessionHelper.OATHDescription = null;
            Impersonation i = null;
            try
            {
                if (SettingsHelper.ProvisionAsUser)
                {
                    i = new Impersonation(SessionHelper.WindowsIdentity);
                }
                var algorithm = (AuthLiteTokenAlgorithm)Enum.Parse(typeof(AuthLiteTokenAlgorithm), AuthLiteOperations.Instance.GetPartitionSetting(SessionHelper.Domain, "DefaultOathAlgorithm", "SHA1") ?? "SHA1", true);
                if (!string.IsNullOrEmpty(SettingsHelper.OathAlgorithm))
                {
                    algorithm = (AuthLiteTokenAlgorithm)Enum.Parse(typeof(AuthLiteTokenAlgorithm), SettingsHelper.OathAlgorithm, true);
                }
                AuthLiteOperations.Instance.ProvisionOATH(SessionHelper.Domain, SessionHelper.Username, i != null ? SessionHelper.Password : null, description, oathSecret, SettingsHelper.OathInterval, 
                        SessionHelper.OATHOnline, 
                        (SettingsHelper.ShouldBootstrap && SessionHelper.OATHOnline), // only bootstrap if it's an online token, else won't really "have a token" yet
                        algorithm
                        );
                    TempData["success"] = "OATH token added successfully.";
                    return RedirectToAction("Index", "Manager");
                }
            catch (Win32Exception e)
            {
                string detail = "";
                switch (e.NativeErrorCode)
                {
                    case 1326:
                        detail = ": Wrong password?";
                        break;
                }
                TempData["error"] = string.Format("Could not use logon token. [Error {0}, {1}{2}]", e.ErrorCode, e.NativeErrorCode, detail);
                return RedirectToAction("Index", "Manager");
            }
            catch (ProvisionTokenException e)
            {
                TempData["error"] = string.Format("Could not assign the token. [Error {0}]", e.Message);
                return RedirectToAction("Index", "Manager");
            }
            catch (Exception e)
            {
                TempData["error"] = string.Format("Could not assign the token. [Error {0}]", e.Message);
                return RedirectToAction("Index", "Manager");
            }
            finally
            {
                if (i != null) i.Dispose();
                i = null;
            }

        }
		/// <summary>
		/// Display the UpdatePin screen
		/// </summary>
		/// <returns></returns>
		public ActionResult UpdatePin()
		{
            UpdatePinFormModel m = new UpdatePinFormModel();
			return View("UpdatePin", m);
		}

		/// <summary>
		/// Handle UpdatePin screen form submission
		/// </summary>
		/// <param name="m"></param>
		/// <returns></returns>
		[HttpPost]
		[ValidateAntiForgeryToken]
		public ActionResult UpdatePin(UpdatePinFormModel m)
		{

			if (!ModelState.IsValid)
			{
				return View(m);
			}
            Impersonation i = null;
            try
            {
                if (SettingsHelper.ProvisionAsUser)
                {
                    i = new Impersonation(SessionHelper.WindowsIdentity);
                }
                    try
                    {
                        AuthLiteOperations.Instance.SetUserPin(SessionHelper.Domain, SessionHelper.Username, m.Pin.Pin);
                    }
                     catch(Win32Exception e)
                    {
                        ModelState.AddModelError("SetUserPinFailed", string.Format("Could not set the user PIN. [Win32 Error {0}, {1}]", e.ErrorCode, e.NativeErrorCode));
				        return View(m);
                    }
                    catch(Exception e)
                    {
                        ModelState.AddModelError("SetUserPinFailed", string.Format("Could not set the user PIN. [General Error {0}]", e.Message));
                        return View(m);
                    }
                }
            catch (Win32Exception e)
            {
                ModelState.AddModelError("ImpersonationFailed", string.Format("Could not use logon token. [Error {0}, {1}]", e.ErrorCode, e.NativeErrorCode));
                return View(m);
            }
            catch (Exception e)
            {
                ModelState.AddModelError("GeneralError", string.Format("[General Error {0}]", e.Message));
                return View(m);
            }
            finally
            {
                if (i != null) i.Dispose();
                i = null;
            }

            TempData["success"] = "User PIN set successfully.";			

			return RedirectToAction("Index", "Manager");
		}
		
		/// <summary>
		/// Display the ViewTokens screen
		/// </summary>
		/// <returns></returns>
		public ActionResult ViewTokens()
		{

			ViewTokensModel m = new ViewTokensModel();
            Impersonation i = null;
            try
            {
                if (SettingsHelper.ProvisionAsUser)
                {
                    i = new Impersonation(SessionHelper.WindowsIdentity);
                }
                    try
                    {
                        m.Tokens = AuthLiteOperations.Instance.ListTokens(SessionHelper.Domain, SessionHelper.Username);
                    }
                    catch (ListTokensException e)
                    {
                        ModelState.AddModelError("ListTokensFailed", string.Format("Could not list tokens. [Error {0}]", e.Message));
                        return View(m);
                    }
                    catch (Win32Exception e)
                    {
                        ModelState.AddModelError("ListTokensFailed", string.Format("Could not list tokens. [Error {0}, {1}]", e.ErrorCode, e.NativeErrorCode));
                        return View(m);
                    }
                    catch (Exception e)
                    {
                        ModelState.AddModelError("GeneralError", string.Format("[General Error {0}]", e.Message));
                        return View(m);
                    }
                }
            catch (Win32Exception e)
            {
                ModelState.AddModelError("ImpersonationFailed", string.Format("Could not use logon token. [Error {0}, {1}]", e.ErrorCode, e.NativeErrorCode));
                return View(m);
            }
            catch (Exception e)
            {
                ModelState.AddModelError("GeneralError", string.Format("[General Error {0}]", e.Message));
                return View(m);
            }
            finally
            {
                if (i != null) i.Dispose();
                i = null;
            }

			return View(m);
		}

		/// <summary>
		/// Disables the user-specified Token after client-side confirmation
		/// </summary>
		/// <param name="m"></param>
		/// <param name="tokenId"></param>
		/// <returns></returns>
		[HttpPost]
		[ValidateAntiForgeryToken]
		public ActionResult DisableToken(ViewTokensModel m, string tokenId)
		{

			/* 
				NOTE
				tokenId will be the value contained in the hidden field matching the submit button pressed
				This value is automatically mapped by MVC according to the string name of the Html.Hidden() helper in the ViewTokens view.
				If the name is changed there, the DisableToken parameter name must also be changed.
			*/
            Impersonation i = null;
            if (!ModelState.IsValid)
            {
                return View(m);
            }

            try
            {
                if (SettingsHelper.ProvisionAsUser)
                {
                    i = new Impersonation(SessionHelper.WindowsIdentity);
                }
                    try
                    {
                        AuthLiteOperations.Instance.DisableToken(SessionHelper.Domain, SessionHelper.Username, tokenId);
                    }
                    catch (DisableTokenException e)
                    {
                        ModelState.AddModelError("DisableTokenFailed", string.Format("Could not disable token. [Error {0}]", e.Message));
                        return View(m);
                    }
                    catch (Win32Exception e)
                    {
                        ModelState.AddModelError("DisableTokenFailed", string.Format("Could not disable token. [Error {0}, {1}]", e.ErrorCode, e.NativeErrorCode));
                        return View(m);
                    }
                    catch (Exception e)
                    {
                        ModelState.AddModelError("GeneralError", string.Format("[General Error {0}]", e.Message));
                        return View(m);
                    }
                }
            catch (Win32Exception e)
            {
                ModelState.AddModelError("ImpersonationFailed", string.Format("Could not use logon token. [Error {0}, {1}]", e.ErrorCode, e.NativeErrorCode));
                return View(m);
            }
            catch (Exception e)
            {
                ModelState.AddModelError("GeneralError", string.Format("[General Error {0}]", e.Message));
                return View(m);
            }

            finally
            {
                if (i != null) i.Dispose();
                i = null;
            }

			return RedirectToAction("ViewTokens", m); // TODO Temporary
		}


	}
}